include(CustomCommand_flex)
include(CustomCommand_bison)
include(CustomCommand_opensg)
include(CustomCommand_qt)
include(CustomCommand_shader)

function (RELATIVE_GLOB VARIABLE DIR FILTER)
    set (_TMP_LIST "")
    if ("${DIR}" STREQUAL ".")
        set (DIR "")
    else ("${DIR}" STREQUAL ".")
        set (DIR "${DIR}/")
    endif ("${DIR}" STREQUAL ".")

	get_filename_component(GLOBDIR "${CMAKE_CURRENT_SOURCE_DIR}/${DIR}" ABSOLUTE)
    
    foreach(_TMP_FILTER ${FILTER})
	
        file(GLOB _TMP_VARIABLE "${GLOBDIR}/${_TMP_FILTER}")
		
        foreach(_TMP_FILE ${_TMP_VARIABLE})
            get_filename_component (_TMP_FILE_NAME ${_TMP_FILE} NAME)
			list (APPEND _TMP_LIST "${DIR}${_TMP_FILE_NAME}")
        endforeach(_TMP_FILE ${_TMP_VARIABLE})
		
    endforeach(_TMP_FILTER ${FILTER})
    set (${VARIABLE} "${_TMP_LIST}" PARENT_SCOPE)
endfunction (RELATIVE_GLOB)

#
# Glob and sort sources

macro (SOURCE_GLOB DIRS)
    set (CPP_SOURCE "")
    set (SHADER_SOURCE "")
    set (OPENSG_SOURCE "")
    set (FLEX_SOURCE "")
    set (BISON_SOURCE "")
    set (UI_SOURCE "")
	set (OSG_HEADERS "")
    
    foreach(_DIR ${DIRS})
    
        # Flex
        RELATIVE_GLOB (_TMP_FLEX_SOURCE ${_DIR} "*.lpp")
        if (_TMP_FLEX_SOURCE)
            set (FLEX_SOURCE ${FLEX_SOURCE} ${_TMP_FLEX_SOURCE})
        endif (_TMP_FLEX_SOURCE)
        
        # Bison
        RELATIVE_GLOB (_TMP_BISON_SOURCE ${_DIR} "*.y")
        if (_TMP_BISON_SOURCE)
            set (BISON_SOURCE ${BISON_SOURCE} ${_TMP_BISON_SOURCE})
        endif (_TMP_BISON_SOURCE)
	
        # OpenSG
        RELATIVE_GLOB (_TMP_OPENSG_SOURCE ${_DIR} "*.fcd")
        if (_TMP_OPENSG_SOURCE)
            set (OPENSG_SOURCE ${OPENSG_SOURCE} ${_TMP_OPENSG_SOURCE})
        endif (_TMP_OPENSG_SOURCE)
        
        # Ui
        RELATIVE_GLOB (_TMP_UI_SOURCE ${_DIR} "*.ui")
        if (_TMP_UI_SOURCE)
            set (UI_SOURCE ${UI_SOURCE} ${_TMP_UI_SOURCE})
        endif (_TMP_UI_SOURCE)
        
        # Shaders
        RELATIVE_GLOB (_TMP_SHADER_SOURCE ${_DIR} "*.vp;*.fp;*.gp")
        if (_TMP_SHADER_SOURCE)
            set (SHADER_SOURCE ${SHADER_SOURCE} ${_TMP_SHADER_SOURCE})
        endif (_TMP_SHADER_SOURCE)
        
        # Cpp and headers
        RELATIVE_GLOB (_TMP_CPP_SOURCE ${_DIR} "OSG*.cpp")
        RELATIVE_GLOB (_TMP_H_SOURCE ${_DIR} "OSG*.h;OSG*.inl")
        if (_TMP_CPP_SOURCE)
            set (CPP_SOURCE ${CPP_SOURCE} ${_TMP_CPP_SOURCE})
        endif (_TMP_CPP_SOURCE)
        if (_TMP_H_SOURCE)
            set (CPP_SOURCE ${CPP_SOURCE} ${_TMP_H_SOURCE})
			set (OSG_HEADERS ${OSG_HEADERS} ${_TMP_H_SOURCE})
        endif (_TMP_H_SOURCE)
        
    endforeach(_DIR ${DIRS})
endmacro (SOURCE_GLOB)

macro (INSTALL_HEADERS)
	set (OSG_HEADERS_TARGETS "")
	if (OSG_HEADERS)
		foreach(_HEADER ${OSG_HEADERS})
			get_filename_component (_TMP_FILE_NAME ${_HEADER} NAME)
			set (_TMP_BIN_DEST "${OPENSG_DIR}/include/OpenSG/${_TMP_FILE_NAME}")
			get_filename_component(_TMP_SOURCE_HEADER "${CMAKE_CURRENT_SOURCE_DIR}/${_HEADER}" ABSOLUTE)
			#set (_TMP_SOURCE_HEADER "${CMAKE_CURRENT_SOURCE_DIR}/${_HEADER}")
			add_custom_command (
				OUTPUT ${_TMP_BIN_DEST}
				COMMAND ${CMAKE_COMMAND} -E make_directory "${OPENSG_DIR}/include/OpenSG"
				COMMAND ${CMAKE_COMMAND} -E copy ${_TMP_SOURCE_HEADER} ${_TMP_BIN_DEST}
				MAIN_DEPENDENCY ${_TMP_SOURCE_HEADER})
			list (APPEND OSG_HEADERS_TARGETS ${_TMP_BIN_DEST})
			set_property(SOURCE ${_TMP_BIN_DEST} PROPERTY GENERATED TRUE)
				
		endforeach(_HEADER ${OSG_HEADERS})
	endif (OSG_HEADERS)
endmacro (INSTALL_HEADERS)

#
# BUILD_CUSTOM
# Used internally in this file for testing if any of the non c++ source variables have been set, and if so: builds the sources depending on the custom rules set up.
# Will make sure to set dependencies correctly and also sort the sorces into an IDE solution display.

macro (BUILD_CUSTOM NAME MODULE_FOLDER)
    set (_TMP_CPP_SOURCE "")
	
    if (SHADER_SOURCE)
        BUILD_SHADERS (_TMP_SHADER_GENERATED "${OPENSG_BUILD}/${MODULE_FOLDER}/${NAME}" "${SHADER_SOURCE}")
        source_group(Shader FILES ${SHADER_SOURCE})
        set (_TMP_CPP_SOURCE ${_TMP_CPP_SOURCE} "${_TMP_SHADER_GENERATED}")
    endif (SHADER_SOURCE)

    if (OPENSG_SOURCE)
        BUILD_OPENSG (_TMP_OPENSG_GENERATED "${OPENSG_BUILD}/${MODULE_FOLDER}/${NAME}" "${OPENSG_SOURCE}")
        source_group(OpenSG FILES ${OPENSG_SOURCE})
        set (_TMP_CPP_SOURCE ${_TMP_CPP_SOURCE} "${_TMP_OPENSG_GENERATED}")
    endif (OPENSG_SOURCE)

    if (FLEX_SOURCE)
        BUILD_FLEX (_TMP_FLEX_GENERATED "${OPENSG_BUILD}/${MODULE_FOLDER}/${NAME}" "${FLEX_SOURCE}")
        source_group(Flex FILES ${FLEX_SOURCE})
        set (_TMP_CPP_SOURCE ${_TMP_CPP_SOURCE} "${_TMP_FLEX_GENERATED}")
    endif (FLEX_SOURCE)

    if (BISON_SOURCE)
        BUILD_BISON (_TMP_BISON_GENERATED "${OPENSG_BUILD}/${MODULE_FOLDER}/${NAME}" "${BISON_SOURCE}")
        source_group(Bison FILES ${BISON_SOURCE})
        set (_TMP_CPP_SOURCE ${_TMP_CPP_SOURCE} "${_TMP_BISON_GENERATED}")
    endif (BISON_SOURCE)

    if (UI_SOURCE)
        BUILD_UI (_TMP_UI_GENERATED "${OPENSG_BUILD}/${MODULE_FOLDER}/${NAME}" "${UI_SOURCE}")
        source_group(QT_Ui FILES ${UI_SOURCE})
        set (_TMP_CPP_SOURCE ${_TMP_CPP_SOURCE} "${_TMP_UI_GENERATED}")
    endif (UI_SOURCE)
    
    if (_TMP_CPP_SOURCE)
		foreach(_SOURCE ${_TMP_CPP_SOURCE})
			set_property(SOURCE ${_SOURCE} PROPERTY GENERATED TRUE)
		endforeach(_SOURCE ${_TMP_CPP_SOURCE})
	
        set (CPP_SOURCE ${CPP_SOURCE} ${_TMP_CPP_SOURCE})
        # Sort generated files into a VS source folder
        source_group(Generated FILES ${_TMP_CPP_SOURCE})
        # Add build fodler as include, so we can find the generated .h files.
        include_directories ("${OPENSG_BUILD}/${MODULE_FOLDER}/${NAME}")
    endif (_TMP_CPP_SOURCE)

endmacro (BUILD_CUSTOM)

#
# CREATION_INIT
# Inits flags and variables needed to create a library or executable.

macro (CREATION_INIT)
    string (REPLACE ";" " " _TMP_CCFLAGS "${CCFLAGS}")

    set (CMAKE_CXX_FLAGS ${_TMP_CCFLAGS})
    set (CMAKE_CXX_FLAGS_DEBUG ${_TMP_CCFLAGS})
    set (CMAKE_CXX_FLAGS_RELEASE ${_TMP_CCFLAGS})

    # Need this as people uses <> instad on "" on local includes...
    include_directories (${CMAKE_CURRENT_SOURCE_DIR})
	
endmacro (CREATION_INIT)

#
# CREATION_DEPEND_AND_LINK
# Adds linkage and dependencies to a library or executable

macro (CREATION_DEPEND_AND_LINK TARGET_NAME)
    if (OSG_HEADERS_TARGETS)
        add_dependencies(${TARGET_NAME} ${OSG_HEADERS_TARGETS})
    endif (OSG_HEADERS_TARGETS)

    if (LIBRARIES)
        target_link_libraries (${TARGET_NAME} ${LIBRARIES})
    endif (LIBRARIES)
    
    if (LINK_DEPENDENCIES)
        target_link_libraries (${TARGET_NAME} ${LINK_DEPENDENCIES})
    endif (LINK_DEPENDENCIES)

endmacro (CREATION_DEPEND_AND_LINK)

#
# CREATION_FINALIZE
# Sets flags and properties to finalize a creation of a library or executable

macro (CREATION_FINALIZE NAME TARGET_NAME MODULE_FOLDER OUTPUT_NAME)
    string (REPLACE ";" " " _TMP_LINK_FLAGS "${LINK_FLAGS}")
    
    set_property(TARGET ${TARGET_NAME} PROPERTY OUTPUT_NAME "${OUTPUT_NAME}")
    set_property(TARGET ${TARGET_NAME} PROPERTY LINK_FLAGS ${_TMP_LINK_FLAGS})
    set_property(TARGET ${TARGET_NAME} PROPERTY FOLDER "${MODULE_FOLDER}")
    set_property(TARGET ${TARGET_NAME} PROPERTY PROJECT_LABEL "${NAME}")
	set_property(TARGET ${TARGET_NAME} PROPERTY LINKER_LANGUAGE CXX)
	
endmacro (CREATION_FINALIZE)

#
# CREATE_SHARED_LIBRARY
# Create a shared library

macro (CREATE_SHARED_LIBRARY NAME TARGET_NAME MODULE_FOLDER OUTPUT_NAME)
    CREATION_INIT()

    BUILD_CUSTOM (${NAME} ${MODULE_FOLDER})
	INSTALL_HEADERS ()
	
    add_library (${TARGET_NAME} SHARED ${CPP_SOURCE})
    
    set_property(TARGET ${TARGET_NAME} PROPERTY RUNTIME_OUTPUT_DIRECTORY "${OPENSG_DIR}/lib")
    set_property(TARGET ${TARGET_NAME} PROPERTY RUNTIME_OUTPUT_DIRECTORY_RELEASE "${OPENSG_DIR}/lib")
    set_property(TARGET ${TARGET_NAME} PROPERTY RUNTIME_OUTPUT_DIRECTORY_DEBUG "${OPENSG_DIR}/lib")
    set_property(TARGET ${TARGET_NAME} PROPERTY ARCHIVE_OUTPUT_DIRECTORY "${OPENSG_DIR}/lib")
    set_property(TARGET ${TARGET_NAME} PROPERTY ARCHIVE_OUTPUT_DIRECTORY_RELEASE "${OPENSG_DIR}/lib")
    set_property(TARGET ${TARGET_NAME} PROPERTY ARCHIVE_OUTPUT_DIRECTORY_DEBUG "${OPENSG_DIR}/lib")
	
	
    CREATION_DEPEND_AND_LINK(${TARGET_NAME})

    CREATION_FINALIZE(${NAME} ${TARGET_NAME} ${MODULE_FOLDER} ${OUTPUT_NAME})
    
endmacro (CREATE_SHARED_LIBRARY)

#
# CREATE_EXECUTABLE
# Create an executable

macro (CREATE_EXECUTABLE NAME TARGET_NAME MODULE_FOLDER OUTPUT_NAME)
    CREATION_INIT()

    BUILD_CUSTOM (${NAME} ${MODULE_FOLDER})
	INSTALL_HEADERS ()
	
    add_executable (${TARGET_NAME} ${CPP_SOURCE})
    
    CREATION_DEPEND_AND_LINK(${TARGET_NAME})

    set_property(TARGET ${TARGET_NAME} PROPERTY RUNTIME_OUTPUT_DIRECTORY "${OPENSG_DIR}/lib")
    set_property(TARGET ${TARGET_NAME} PROPERTY RUNTIME_OUTPUT_DIRECTORY_RELEASE "${OPENSG_DIR}/lib")
    set_property(TARGET ${TARGET_NAME} PROPERTY RUNTIME_OUTPUT_DIRECTORY_DEBUG "${OPENSG_DIR}/lib")

    CREATION_FINALIZE(${NAME} ${TARGET_NAME} ${MODULE_FOLDER} ${OUTPUT_NAME})
    
endmacro (CREATE_EXECUTABLE)

