
function(BUILD_UI OUTPUT DESTDIR SOURCES)
    set (_TMP_OUTPUT "")
    
    foreach(_SOURCE ${SOURCES})
	
        get_filename_component (_TMP_FILE_PATH ${_SOURCE} PATH)
        set(_TMP_FILE_PATH "${DESTDIR}/${_TMP_FILE_PATH}")
        get_filename_component (_TMP_FILE_NAME ${_SOURCE} NAME_WE)
        
        set (UIC "${QT_DIR}/${PLATFORM_DIR}/bin/uic")

        set (_TMP_BUILD_H_DEST "${OPENSG_DIR}/include/OpenSG/${_TMP_FILE_NAME}.h")
        add_custom_command (
            OUTPUT ${_TMP_BUILD_H_DEST}
            COMMAND ${CMAKE_COMMAND} -E make_directory "${_TMP_FILE_PATH}"
            COMMAND ${UIC} -o ${_TMP_BUILD_H_DEST} ${_SOURCE}
            DEPENDS ${_SOURCE}
            WORKING_DIRECTORY ${CMAKE_CURRENT_SOURCE_DIR})

        set (_TMP_BUILD_MOC_CPP_DEST "${_TMP_FILE_PATH}/moc_${_TMP_FILE_NAME}.cpp")
        list (APPEND _TMP_OUTPUT ${_TMP_BUILD_MOC_CPP_DEST})
        add_custom_command (
            OUTPUT ${_TMP_BUILD_MOC_CPP_DEST}
            COMMAND ${CMAKE_COMMAND} -E make_directory "${_TMP_FILE_PATH}"
            COMMAND ${QT_DIR}/${PLATFORM_DIR}/bin/moc -DQT_OPENGL_SUPPORT -DQT_DLL  ${_TMP_BUILD_H_DEST} -o ${_TMP_BUILD_MOC_CPP_DEST}
            DEPENDS ${_TMP_BUILD_H_DEST}
            WORKING_DIRECTORY ${CMAKE_CURRENT_SOURCE_DIR})
            
		set (_TMP_BUILD_CPP_DEST "${_TMP_FILE_PATH}/${_TMP_FILE_NAME}.cpp")
		list (APPEND _TMP_OUTPUT ${_TMP_BUILD_CPP_DEST})
		add_custom_command (
			OUTPUT ${_TMP_BUILD_CPP_DEST}
			COMMAND ${CMAKE_COMMAND} -E make_directory "${_TMP_FILE_PATH}"
			COMMAND ${UIC} -o ${_TMP_BUILD_CPP_DEST} -impl ${_TMP_BUILD_H_DEST} ${_SOURCE}
			DEPENDS ${_SOURCE} ${_TMP_BUILD_H_DEST}
			WORKING_DIRECTORY ${CMAKE_CURRENT_SOURCE_DIR})
            
    endforeach(_SOURCE ${SOURCES})
    
    set (${OUTPUT} ${_TMP_OUTPUT} PARENT_SCOPE)
endfunction(BUILD_UI)

function(BUILD_MOC OUTPUT DESTDIR SOURCES)
    set (_TMP_OUTPUT "")

	set (_TMP_MOC_DEFINES "")
	get_directory_property(_TMP_CPP_DEFINES COMPILE_DEFINITIONS)
    foreach(_DEFINE ${_TMP_CPP_DEFINES})
		set (_TMP_MOC_DEFINES "${_TMP_MOC_DEFINES} -D${_DEFINE}")
    endforeach(_DEFINE ${_TMP_CPP_DEFINES})
    
    foreach(_SOURCE ${SOURCES})
        get_filename_component (_TMP_FILE_PATH ${_SOURCE} PATH)
        set(_TMP_FILE_PATH "${DESTDIR}/${_TMP_FILE_PATH}")
        get_filename_component (_TMP_FILE_NAME ${_SOURCE} NAME_WE)
        get_filename_component (_TMP_FILE_EXT ${_SOURCE} EXT)
        
        if (_TMP_FILE_EXT STREQUAL ".cpp")
            set (_TMP_BUILD_CPP_DEST "${_TMP_FILE_PATH}/moc_cpp_${_TMP_FILE_NAME}.cpp")
        else (_TMP_FILE_EXT STREQUAL ".cpp")
            set (_TMP_BUILD_CPP_DEST "${_TMP_FILE_PATH}/moc_${_TMP_FILE_NAME}.cpp")
        endif (_TMP_FILE_EXT STREQUAL ".cpp")

        list (APPEND _TMP_OUTPUT ${_TMP_BUILD_CPP_DEST})
        if (_TMP_FILE_EXT STREQUAL ".cpp")
			# Some weird hack in QtPropertyBrowser uses moc'ed cpp files.
			# But if we have a main dependency to a cpp file, then the cpp file will only get moc'ed, not compiled.
			# So we use just a dependency instead, and not a main dependency.
			add_custom_command (
				OUTPUT ${_TMP_BUILD_CPP_DEST}
				COMMAND ${CMAKE_COMMAND} -E make_directory "${_TMP_FILE_PATH}"
				COMMAND ${QT_DIR}/${PLATFORM_DIR}/bin/moc ${_TMP_MOC_DEFINES}  ${_SOURCE} -o ${_TMP_BUILD_CPP_DEST}
				DEPENDS ${_SOURCE}
				WORKING_DIRECTORY ${CMAKE_CURRENT_SOURCE_DIR})
        else (_TMP_FILE_EXT STREQUAL ".cpp")
			add_custom_command (
				OUTPUT ${_TMP_BUILD_CPP_DEST}
				COMMAND ${CMAKE_COMMAND} -E make_directory "${_TMP_FILE_PATH}"
				COMMAND ${QT_DIR}/${PLATFORM_DIR}/bin/moc ${_TMP_MOC_DEFINES}  ${_SOURCE} -o ${_TMP_BUILD_CPP_DEST}
				DEPENDS ${_SOURCE}
				WORKING_DIRECTORY ${CMAKE_CURRENT_SOURCE_DIR})
        endif (_TMP_FILE_EXT STREQUAL ".cpp")
    endforeach(_SOURCE ${SOURCES})
    
    set (${OUTPUT} ${_TMP_OUTPUT} PARENT_SCOPE)
endfunction(BUILD_MOC)

