/*
 * Copyright (c) 2008
 * Evan Teran
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appears in all copies and that both the
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the same name not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission. We make no representations about the
 * suitability this software for any purpose. It is provided "as is"
 * without express or implied warranty.
 */

#ifndef _OSGUINT128_H_
#define _OSGUINT128_H_

#include <stdexcept>
#include <string>
#include <iostream>

#include <OSGConfig.h>
#include <OSGBaseTypes.h>

OSG_BEGIN_NAMESPACE

class OSG_BASE_DLLMAPPING UInt128
{
    typedef UInt64 base_type;

public:

    UInt128(void);
    UInt128(Int32 value);
    UInt128(UInt32 value);
    UInt128(const UInt128 &value);

    UInt128 &operator=(const UInt128 &other);
    bool operator==(const UInt128 &o) const;

    /*
    bool operator==(const bool &o) const;
    */
    bool operator!=(const UInt128 &o) const;
    /*
    bool operator!=(const Int64 &o) const;
    bool operator<(const UInt128 &o) const;
    bool operator>=(const UInt128 &o) const;
    bool operator!() const;
    */
    //operator bool() const;

    UInt128 operator~() const;

    UInt128 operator>>(const Int32 &shift) const;
    UInt128 operator<<(const Int32 &shift) const;

    UInt128 operator&(const UInt128 &b);
    //UInt128 operator&(const Int32 &b);

    UInt128 operator|(const UInt128 &b);
    UInt128 operator^(const UInt128 &b);

    UInt128 &operator|=(const UInt128 &b);
    UInt128 &operator&=(const UInt128 &b);
    UInt128 &operator^=(const UInt128 &b);

    UInt128 &operator<<=(const Int32 &shift);
    UInt128 &operator>>=(const Int32 &shift);

public:

    static const unsigned int size = (sizeof(base_type) + sizeof(base_type)) * 8;

    base_type lo;
    base_type hi;
};

UInt128 operator&(const UInt128 &a, const UInt128 &b)
{
    UInt128 result(a);
    result &= b;
    return result;
}

std::ostream &operator<<(std::ostream &o, const UInt128 &n)
{
    UInt128 s = 1;
    for(Int32 i=127;i>=0;--i)
        o << (((n & (s << i)) == 0) ? 0 : 1);
    return o;
}

template <>
struct TypeTraits<UInt128> : public TypeTraitsBase
{
    //typedef       Real32             RealReturnType;


    static const  bool               IsPOD       = true;
    static const  MathTypeProperties MathProp    = IntValue;

#ifdef OSG_LONGLONG_HAS_LL
    /*
    static const  UInt64             BitsSet     = 0xFFFFFFFFFFFFFFFFLL;
    static const  UInt64             BitsClear   = 0x0000000000000000LL;

    static const  UInt64             Zero        = 0x0000000000000000LL;
    static const  UInt64             One         = 0x0000000000000001LL;
    */
    static const  UInt128            Zero;
    static const  UInt128            One;
#else
    /*
    static const  UInt64             BitsSet     = 0xFFFFFFFFFFFFFFFF;
    static const  UInt64             BitsClear   = 0x0000000000000000;
    */
    static const  UInt128            Zero;
    static const  UInt128            One;
#endif
 


    static        UInt128             getZeroElement(void)
    {
        return 0;
    }

    static        UInt128             getOneElement (void)
    {
        return 1;
    }

/*
    static        UInt64             getMax        (void)         
    {
        return std::numeric_limits<UInt64>::max();
    }

    static        UInt64             getMin        (void)
    {
        return std::numeric_limits<UInt64>::min();
    }


    static Real32      getFraction  (UInt64 val)
    {
        return (Real32(val) / Real32(getMax()));
    };

    static UInt64      getPortion   (Real32 val)
    {
        return UInt64(val * Real32(getMax()));
    };


    static UInt64      getFromString(const Char8 *szString)
    {
        if(szString != NULL)
        {
#ifndef WIN32
#ifdef __sgi
            // on my sgi machine strtoull is defined in the header
            // but not in the lib. amz
            UInt64 r = 0;
            sscanf(szString, "%llu", &r);
            return r;
#else
            return UInt64(strtoull(szString, NULL, 0));
#endif
#else
            return _atoi64(szString);
#endif
        }
        else
        {
            return getZeroElement();
        }
    }

    static std::string putToString  (const UInt64 val)
    {
        Char8 buffer[25];
#ifdef OSG_LINUX_TYPES
        sprintf(buffer, "%"PRIu64, val );
#else
        sprintf(buffer, "%llu", val);
#endif
        return std::string(buffer);
    }
*/
};

OSG_END_NAMESPACE

#define OSGUINT128_HEADER_CVSID "@(#)$Id: $"

#endif /* _OSGUINT128_H_ */
