#ifndef _GROUPIBPIPELINE_H_
#define _GROUPIBPIPELINE_H_
#pragma once

#include <string>
#include <memory>
#include <OSGBase.h>
#include <OSGBaseTypes.h>
#include <OSGGroupSockConnection.h>
#include <OSGTime.h>
#include <OSGStreamSocket.h>
#include <OSGGroupSockPipeline.h>
#include <OSGDgramQueue.h>

#include "OSGPointIBConnection.h"

OSG_BEGIN_NAMESPACE

class PointIBPipeline;

class OSG_BASE_DLLMAPPING GroupIBPipeline : public GroupSockConnection
{
  public:
    GroupIBPipeline(void);
    virtual ~GroupIBPipeline(void);
    virtual const ConnectionType *getType(void) const;

    virtual Channel connectPoint(const std::string &address, Time timeout = -1);
    virtual void disconnect(Channel channel);
    virtual void disconnect(const std::vector<Channel> &channels);
    virtual Channel acceptPoint(Time timeout = -1);
    virtual std::string bind(const std::string &interf);

    virtual void setParams(const std::string &params);
    virtual Channel selectChannel(Time timeout = -1) throw(ReadError);

    virtual int getAvailable(void);
    virtual int getAvailable(Channel channel);
    virtual int waitAvailable(Channel channel);

    virtual Channel selectSingleChannel(Channel channel,
                                        Time timeout = -1) throw(ReadError);
    virtual Int32
    selectSingleChannelAndGetAvaiable(Channel channel,
                                      Time timeout) throw(ReadError);
    virtual bool hasAvaiable(Channel channel, Int32 timeout) throw(ReadError);

    virtual std::vector<Channel>
    selectChannels(Time timeout = -1) throw(ReadError);

    virtual void flush(void);

    virtual void setDelay(bool s);

    virtual SizeT read(MemoryHandle mem, SizeT size);
    virtual SizeT read(Channel channel, MemoryHandle mem, SizeT size);

    virtual void readBuffer(void);

    virtual void write(const MemoryHandle mem, SizeT size, bool force_unicast = false);
    virtual void write(Channel channel, const MemoryHandle mem, SizeT size);

    virtual void writeBuffer(void);
    virtual bool wait(Time timeout) throw(ReadError);
    virtual void signal(void) throw(WriteError);
    static GroupConnection *create(void);

  protected:
  private:
    typedef std::vector<UInt8> DataBuffer;
    typedef std::vector<PointIBConnection *> PointConnections;
    typedef GroupSockConnection Inherited;

    void initialize();
    void initialize(
        PointIBConnection *target, const PointConnections &connections,
        const std::vector<std::string> &addresses);

    friend class PointIBPipeline;

    ChannelIndex _readIndex;
    PointConnections _connections;
    PointConnections _next;
    bool _initialized;
    int _maxNodeCount;

    static ConnectionType _type;

    // prohibit default functions (move to 'public' if you need one)
    GroupIBPipeline(const GroupIBPipeline &source);
    GroupIBPipeline &operator=(const GroupIBPipeline &source);
};

typedef GroupIBPipeline *GroupIBPipelineP;

OSG_END_NAMESPACE

#define OSG__HEADER_CVSID "@(#)$Id: $"
#endif // _GROUPIBPIPELINE_H_
