#ifndef _POINTIBCONNECTION_H_
#define _POINTIBCONNECTION_H_
#pragma once

#include <sstream>
#include <string>
#include <queue>
#include <stdexcept>

#include <OSGBase.h>
#include <OSGBaseTypes.h>
#include <OSGPointConnection.h>
#include <OSGPointSockConnection.h>
#include <OSGTime.h>
#include <OSGStreamSocket.h>

OSG_BEGIN_NAMESPACE

class IBImplementation;
class OSG_BASE_DLLMAPPING PointIBConnection : public PointSockConnection
{
  public:
    PointIBConnection(StreamSocket acceptSocket = StreamSocket());
    virtual ~PointIBConnection(void);

    virtual const ConnectionType *getType(void) const;

    virtual Channel connectPoint(const std::string &address, Time timeout = -1);
    virtual Channel connectGroup(const std::string &address, Time timeout = -1);
    virtual void disconnect(void);
    virtual Channel acceptGroup(Time timeout = -1);
    virtual Channel acceptPoint(Time timeout = -1);
    virtual std::string bind(const std::string &interf);
    virtual bool wait(Time timeout) throw(ReadError);
    virtual void signal(void) throw(WriteError);
    virtual Channel selectChannel(Time timeout = -1) throw(ReadError);
    virtual int getAvailable(void);
    virtual void setCork(bool s);
    virtual void setDelay(bool s);
    virtual SizeT read(MemoryHandle mem, SizeT size);
    virtual void write(const MemoryHandle mem, SizeT size);
    virtual int waitAvailable(void);
    virtual void readBuffer(void);
    virtual void writeBuffer(void);

    static PointConnection *create(void);

  private:
    IBImplementation *_ib;

    static ConnectionType _type;

    typedef PointSockConnection Inherited;

    // prohibit default functions (move to 'public' if you need one)
    PointIBConnection(const PointIBConnection &source);
    PointIBConnection &operator=(const PointIBConnection &source);
};

typedef PointIBConnection *PointIBConnectionP;

OSG_END_NAMESPACE

#define OSG_CONNECTION_HEADER_CVSID "@(#)$Id: $"

#endif /* _POINTIBCONNECTION_H_ */
