//#define MULTI_THREADED
#define THRESHOLD_SUPPORT

#include <iostream>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include "OSGConfig.h"
#include "OSGLog.h"
#include "OSGBaseFunctions.h"
#include "OSGBaseThread.h"
#include "OSGSocketSelection.h"
#include "OSGBinaryMessage.h"
#include "OSGPointIBPipeline.h"
#include "OSGGroupSockPipeline.h"
#include "OSGConnectionType.h"
#include "OSGBinaryMessage.h"

#include "OSGminiz.h"
#include "OSGLZ4.h"

OSG_USING_NAMESPACE

PointIBPipeline::PointIBPipeline()
    : Inherited(), _initialized(false), _initializing(false), _next(), _prev()
{
    _prev = new PointIBConnection();
}

PointIBPipeline::~PointIBPipeline(void)
{
    for (int index = 0; index < _next.size(); ++index)
        delete _next[index];
}

const ConnectionType *PointIBPipeline::getType() const
{
    return &_type;
}

Connection::Channel PointIBPipeline::connectGroup(const std::string &address,
                                                  Time timeout)
{
    _initialized = _initializing = false;
    return Inherited::connectGroup(address,timeout);
}

void PointIBPipeline::disconnect(void)
{
    Inherited::disconnect();
}

Connection::Channel PointIBPipeline::acceptGroup(Time timeout)
{
    _initialized = _initializing = false;
    return Inherited::acceptGroup(timeout);
}

Connection::Channel
PointIBPipeline::selectChannel(Time timeout) throw(ReadError)
{
    try
    {
        if(_pointToPoint)
            return Inherited::selectChannel(timeout);
        if (_initializing)
            return Inherited::selectChannel(timeout);
        if (!_initialized)
            initialize();
        return _prev->selectChannel(timeout);
    }
    catch (SocketError &e)
    {
        throw ReadError(e.what());
    }
    return -1;
}

PointConnection *PointIBPipeline::create(void)
{
    return new PointIBPipeline();
}

SizeT PointIBPipeline::read(MemoryHandle mem, SizeT size)
{
    if(_pointToPoint)
        return Inherited::read(mem, size);

    int len = 0;
    if (_initializing)
        return Inherited::read(mem,size);
    if (!_initialized)
        initialize();
    // read data
    _prev->read(mem, size);
    // send to next in chain
    for(int index=0;index < _next.size();++index)
        _next[index]->write(mem,size);
    return len;
}

void PointIBPipeline::readBuffer()
{
    if(_pointToPoint)
    {
        Inherited::readBuffer();
        return;
    }
    int size;
    int len;

    if (_initializing)
    {
        Inherited::readBuffer();
        return;
    }
    if (!_initialized)
        initialize();
    // read buffer header
    len = _prev->read(&_socketReadBuffer[0], sizeof(SocketBufferHeader));
    if (len == 0)
        throw ReadError("peek got 0 bytes!");

    // read remaining data
    size = osgntohl(
        reinterpret_cast<SocketBufferHeader *>(&_socketReadBuffer[0])->size);
    len = _prev->read(&_socketReadBuffer[sizeof(SocketBufferHeader)], size);
    if (len == 0)
        throw ReadError("read got 0 bytes!");

    for(int index=0;index < _next.size();++index)
        _next[index]->write(&_socketReadBuffer[0], sizeof(SocketBufferHeader) + size);

    readBufBegin()->setDataSize(size);
}

bool PointIBPipeline::wait(Time timeout) throw (ReadError)
{
    return Inherited::wait(timeout);
}

void PointIBPipeline::signal(void) throw(WriteError)
{
    Inherited::signal();
}

void PointIBPipeline::initialize(void)
{
    std::string interf;

    _initializing = true;

    // get local host name doesn't work in all cases we need a ip
    // so we already have a open connection via _socket and we read the ip from
    // there!
    // osgGetHostname(localhost, 255);
    if (!getInterface().empty())
        interf = getInterface();
    else
        interf = _socket.getAddress().getHost(); // _group->getSocketHost();

    // send my own address
    std::string boundTo = _prev->bind(interf.c_str());

    putValue(boundTo);
    flush();

    UInt32 count;
    getValue(count);
    for (int index = 0; index < count; ++index)
    {
        std::string address;
        getValue(address);
        PointIBConnection *next = new PointIBConnection();
        next->connectPoint(address);
        _next.push_back(next);
    }
    _prev->acceptPoint();
    _initialized = true;
    _initializing = false;
}

ConnectionType PointIBPipeline::_type(&PointIBPipeline::create, "IBPipeline");

#ifdef __sgi
#pragma set woff 1174
#endif

#ifdef OSG_LINUX_ICC
#pragma warning(disable : 177)
#endif

namespace
{
static Char8 cvsid_cpp[] = "@(#)$Id: $";
static Char8 cvsid_hpp[] = OSG_GROUPSOCKPIPELINE_HEADER_CVSID;
}
