#ifndef _POINTIBPIPELINE_H_
#define _POINTIBPIPELINE_H_
#pragma once

#include <string>
#include <map>
#include <OSGBase.h>
#include <OSGBaseTypes.h>
#include <OSGTime.h>
#include <OSGDgramSocket.h>
#include <OSGDgram.h>
#include <OSGDgramQueue.h>
#include "OSGThreadPool.h"
#include "OSGPointIBConnection.h"

OSG_BEGIN_NAMESPACE

class OSG_BASE_DLLMAPPING PointIBPipeline : public PointIBConnection
{
  public:
    PointIBPipeline(void);
    virtual ~PointIBPipeline(void);

    virtual const ConnectionType *getType(void) const;

    virtual Channel connectGroup(const std::string &address, Time timeout = -1);
    virtual void disconnect(void);
    virtual Channel acceptGroup(Time timeout = -1);

    virtual Channel selectChannel(Time timeout = -1) throw(ReadError);

    static PointConnection *create(void);

  protected:
    virtual SizeT read(MemoryHandle mem, SizeT size);
    virtual void readBuffer(void);
    virtual bool wait  (Time timeout) throw (ReadError );
    virtual void signal(void) throw(WriteError);

    PointIBConnection *_prev;
    std::vector<PointIBConnection *> _next;
    bool _initialized;
    bool _initializing;

  private:
    void initialize(void);
    static ConnectionType _type;

    typedef PointIBConnection Inherited;

    // prohibit default functions (move to 'public' if you need one)
    PointIBPipeline(const PointIBPipeline &source);
    PointIBPipeline &operator=(const PointIBPipeline &source);
};

typedef PointIBPipeline *PointIBPipelineP;

OSG_END_NAMESPACE

#define OSG_GROUPSOCKPIPELINE_HEADER_CVSID "@(#)$Id: $"

#endif /* _GROUPSOCKPIPELINE_H_ */
