/*---------------------------------------------------------------------------*\
 *                                OpenSG                                     *
 *                                                                           *
 *                                                                           *
 *             Copyright (C) 2000-2002 by the OpenSG Forum                   *
 *                                                                           *
 *                            www.opensg.org                                 *
 *                                                                           *
 *   contact: dirk@opensg.org, gerrit.voss@vossg.org, jbehr@zgdv.de          *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                License                                    *
 *                                                                           *
 * This library is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU Library General Public License as published    *
 * by the Free Software Foundation, version 2.                               *
 *                                                                           *
 * This library is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 * Library General Public License for more details.                          *
 *                                                                           *
 * You should have received a copy of the GNU Library General Public         *
 * License along with this library; if not, write to the Free Software       *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                Changes                                    *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
\*---------------------------------------------------------------------------*/


#ifndef _GROUPSOCKCONNECTION_H_
#define _GROUPSOCKCONNECTION_H_
#pragma once

//---------------------------------------------------------------------------
//  Includes
//---------------------------------------------------------------------------

#include <string>
#include <vector>

#include <OSGBase.h>
#include <OSGBaseTypes.h>
#include <OSGGroupConnection.h>
#include <OSGTime.h>
#include <OSGStreamSocket.h>
#include "OSGThreadPool.h"

OSG_BEGIN_NAMESPACE

class OSG_BASE_DLLMAPPING GroupSockConnection : public GroupConnection
{
    /*==========================  PUBLIC  =================================*/
  public:

    /*---------------------------------------------------------------------*/
    /*! \name                   Constructors                               */
    /*! \{                                                                 */

             GroupSockConnection ( void );
    virtual ~GroupSockConnection ( void );

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                   type info                                  */
    /*! \{                                                                 */

    virtual const ConnectionType *getType (void) const;

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                   connection                                 */
    /*! \{                                                                 */

    virtual Channel     connectPoint(const std::string &address,
                                           Time        timeout=-1 );
    virtual void        disconnect  (      Channel     channel    );
    virtual void        disconnect(const std::vector<Channel> &channels);
    virtual Channel     acceptPoint (      Time        timeout=-1 );
    virtual std::string bind        (const std::string &interf    );

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                   params                                     */
    /*! \{                                                                 */

    virtual void setParams(const std::string &params);

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                   channel handling                           */
    /*! \{                                                                 */

    virtual Channel selectChannel   (Time timeout=-1) throw (ReadError);

    virtual int getAvailable        (void           );
    virtual int getAvailable        (Channel channel);
    virtual int waitAvailable       (Channel channel);

    virtual Channel selectSingleChannel(Channel channel, Time timeout = -1) throw (ReadError);
    virtual Int32 selectSingleChannelAndGetAvaiable(Channel channel, Time timeout) throw (ReadError);
    virtual bool hasAvaiable(Channel channel, Int32 timeout) throw (ReadError);

    virtual std::vector<Channel> selectChannels(Time timeout=-1) throw (ReadError);

    virtual void    flush       (void      );

    virtual void setDelay(bool s);

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                   create                                     */
    /*! \{                                                                 */

    static GroupConnection *create(void);

    /*! \}                                                                 */

    /*=========================  PROTECTED  ===============================*/
  protected:

    /*---------------------------------------------------------------------*/
    /*! \name                   Types                                      */
    /*! \{                                                                 */

    struct SocketBufferHeader
    {
        UInt32 magic;
        UInt32 id;
        UInt32 size;
        UInt32 compressed_size;
    };

    typedef std::vector<UInt8>          DataBuffer;
    typedef std::vector<StreamSocket>   SocketVec;
    typedef std::vector<SocketAddress>  AddressVec;


    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                   IO Implementation                          */
    /*! \{                                                                 */

    virtual SizeT read             (MemoryHandle mem, SizeT size);
    virtual SizeT read             (Channel channel, MemoryHandle mem, SizeT size);

    virtual void getFromChannel        (Channel channel, void         *dst,   UInt32 size);
    virtual void getValueFromChannel   (Channel channel, bool         &value             );
    virtual void getValueFromChannel   (Channel channel, UInt8        &value             );
    virtual void getValueFromChannel   (Channel channel, UInt16       &value             );
    virtual void getValueFromChannel   (Channel channel, UInt32       &value             );
    virtual void getValueFromChannel   (Channel channel, UInt64       &value             );
    virtual void getValueFromChannel   (Channel channel, Int8         &value             );
    virtual void getValueFromChannel   (Channel channel, Int16        &value             );
    virtual void getValueFromChannel   (Channel channel, Int32        &value             );
    virtual void getValueFromChannel   (Channel channel, Int64        &value             );
    virtual void getValueFromChannel   (Channel channel, Real16       &value             );
    virtual void getValueFromChannel   (Channel channel, Real32       &value             );
    virtual void getValueFromChannel   (Channel channel, Real64       &value             );
    virtual void getValueFromChannel   (Channel channel, Real128      &value             );
    virtual void getValueFromChannel   (Channel channel, std::string  &value             );

    virtual void readBuffer       (void);

    virtual void write            (const MemoryHandle mem, SizeT size, bool force_unicast = false);
    virtual void write            (Channel channel, const MemoryHandle mem, SizeT size);

    virtual void putToChannel       (Channel channel, void const   *src,   UInt32 size);
    virtual void putValueToChannel  (Channel channel, const bool         &value       );
    virtual void putValueToChannel  (Channel channel, const UInt8        &value       );
    virtual void putValueToChannel  (Channel channel, const UInt16       &value       );
    virtual void putValueToChannel  (Channel channel, const UInt32       &value       );
    virtual void putValueToChannel  (Channel channel, const UInt64       &value       );
    virtual void putValueToChannel  (Channel channel, const Int8         &value       );
    virtual void putValueToChannel  (Channel channel, const Int16        &value       );
    virtual void putValueToChannel  (Channel channel, const Int32        &value       );
    virtual void putValueToChannel  (Channel channel, const Int64        &value       );
    virtual void putValueToChannel  (Channel channel, const Real16       &value       );
    virtual void putValueToChannel  (Channel channel, const Real32       &value       );
    virtual void putValueToChannel  (Channel channel, const Real64       &value       );
    virtual void putValueToChannel  (Channel channel, const Real128      &value       );
    virtual void putValueToChannel  (Channel channel, const std::string  &value       );

    virtual void writeBuffer      (void);

    static void writeThread       (void *arg);

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                   members                                    */
    /*! \{                                                                 */

    static const UInt32 _ioBufferSize;

    StreamSocket               _acceptSocket;
    SocketVec    _sockets;
    AddressVec   _remoteAddresses;
    ChannelIndex               _readIndex;
    DataBuffer   _socketReadBuffer;
    DataBuffer   _socketWriteBuffer;
    ThreadPool   _tp;
    UInt32       _id;

    /*! \}                                                                 */
    /*---------------------------------------------------------------------*/
    /*! \name                   synchronisation                            */
    /*! \{                                                                 */

    virtual bool wait  (Time timeout) throw (ReadError);
    virtual void signal(void        ) throw (WriteError);

    /*! \}                                                                 */

    /*==========================  PRIVATE  ================================*/
  private:

    /*---------------------------------------------------------------------*/
    /*! \name                   static type                                */
    /*! \{                                                                 */

    static ConnectionType _type;

    /*! \}                                                                 */

    typedef GroupConnection Inherited;

    // prohibit default functions (move to 'public' if you need one)
    GroupSockConnection(const GroupSockConnection &source);
    GroupSockConnection& operator =(const GroupSockConnection &source);
};

//---------------------------------------------------------------------------
//   Exported Types
//---------------------------------------------------------------------------

// class pointer

typedef GroupSockConnection *GroupSockConnectionP;

OSG_END_NAMESPACE

#define OSG_GROUPSOCKCONNECTION_HEADER_CVSID "@(#)$Id: $"

#endif /* _GROUPSOCKCONNECTION_H_ */
