#include <OSGConfig.h>

#include <iostream>

#include <OSGBaseFunctions.h>
#include <OSGFieldContainerFactory.h>
#include <OSGSFSysTypes.h>
#include <OSGNode.h>
#include <OSGDirectionalLight.h>
#include "OSGAction.h"

using namespace OSG;

// derive two NodeCores so that we have something to work on.
// As simple as possible. This is not how to do it right!

class Core1;
typedef FCPtr<NodeCorePtr, Core1> Core1Ptr;

class Core1 : public NodeCore
{
  public:
    static const char *getClassname(void) { return "Core1"; };

    OSG_FIELD_CONTAINER_DECL(Core1Ptr)

    Core1(void) {}
    virtual ~Core1(void) {}

   Action::ResultE enter(Action * action )
   {
        std::cerr << "Core1::enter: " << this << " action " 
             << action << std::endl;
        return Action::Continue;        
   }
   
   Action::ResultE leave(Action * action )
   {
        std::cerr << "Core1::leave: " << this << " action " 
             << action << std::endl;
        return Action::Continue;        
   }
    
  virtual void execBeginEdit(const osg::BitVector&, osg::UInt32, osg::UInt32)
  {
    std::cerr << "Core1::execBeginEdit: " << this << std::endl;
  }
  
  private:

    typedef NodeCore Inherited;

    friend class FieldContainer;

    static FieldContainerType _type;    
};

OSG_FIELD_CONTAINER_INL_DEF(Core1, Core1Ptr)
OSG_FIELD_CONTAINER_DEF(Core1, Core1Ptr)

FieldContainerType Core1::_type(
    "Core1",
    "NodeCore",
    0,
    reinterpret_cast<PrototypeCreateF>(&Core1::createEmpty)
);


// Core2: derived from Core1, to test inheritance following

class Core2;
typedef FCPtr<Core1Ptr, Core2> Core2Ptr;

class Core2 : public Core1
{
  public:
    static const char *getClassname(void) { return "Core2"; };

    OSG_FIELD_CONTAINER_DECL(Core2Ptr)

    Core2(void) {}
    virtual ~Core2(void) {}

  private:

    typedef NodeCore Inherited;

    friend class FieldContainer;

    static FieldContainerType _type;    
};

OSG_FIELD_CONTAINER_INL_DEF(Core2, Core2Ptr)
OSG_FIELD_CONTAINER_DEF(Core2, Core2Ptr)

FieldContainerType Core2::_type(
    "Core2",
    "Core1",
    0,
    reinterpret_cast<PrototypeCreateF>(&Core2::createEmpty)
);





// call a method of a class instance

class Core1Action
{
public:
    Core1Action() {}

    Action::ResultE enter(const NodeCorePtr &core, Action * action)
    { 
        std::cerr << "Core1 enter: " << core << " action "
             << std::hex << action << " for Core1 " << this << std::endl;
        return Action::Continue; 
    }

    Action::ResultE leave(const NodeCorePtr &core, Action * action)
    { 
        std::cerr << "Core1 leave: " << core << " action "
             << std::hex << action << " for Core1 " << this << std::endl;
        return Action::Continue; 
    }
};


// Core2: use simple functions

Action::ResultE Core2Enter(const NodeCorePtr& core, Action * action)
{ 
    std::cerr << "Core2 enter: " << core
         << " action " << std::hex << action << std::endl;

    return Action::Continue; 
}

Action::ResultE Core2Leave(const NodeCorePtr& core, Action * action)
{ 
    std::cerr << "Core2 leave: " << core
         << " action " << std::hex << action << std::endl;

    return Action::Continue; 
}


// traverse only the first node

Action::ResultE firstOnly(const NodeCorePtr& core, Action * action)
{ 
    std::cerr << "Core1 (first only) enter: " << core << " action " << std::hex << action
         << std::endl;

    const NodePtr p = action->getNode( 0 );

    if ( p != NullFC )
        action->addNode( p );

    return Action::Continue; 
}

// default function

Action::ResultE defenter1(const NodeCorePtr& core, Action * action)
{ 
    std::cerr << "Default enter 1 called: " << core
         << " action " << std::hex << action << std::endl;

    return Action::Continue; 
}

Action::ResultE defleave1(const NodeCorePtr& core, Action * action)
{ 
    std::cerr << "Default leave 1 called: " << core
         << " action " << std::hex << action << std::endl;

    return Action::Continue; 
}

Action::ResultE defenter2(const NodeCorePtr& core, Action * action)
{ 
    std::cerr << "Default enter 2 called: " << core
         << " action " << std::hex << action << std::endl;

    return Action::Continue; 
}

Action::ResultE defleave2(const NodeCorePtr& core, Action * action)
{ 
    std::cerr << "Default leave 2 called: " << core
         << " action " << std::hex << action << std::endl;

    return Action::Continue; 
}


Action::ResultE traventer(const NodePtr& node)
{ 
    std::cerr << "traventer called: " << node << std::endl;

    return Action::Continue; 
}

Action::ResultE travleave(const NodePtr& node, Action::ResultE res)
{ 
    std::cerr << "travleave called: " << node << " " << res << std::endl;

    return Action::Continue; 
}

struct TravActor
{
    Action::ResultE enter(const NodePtr& n)
    {
        std::cerr << "TravActor::enter " << this << " Node " << n << std::endl;
        return Action::Continue;
    }
    Action::ResultE leave(const NodePtr& n, Action::ResultE res)
    {
        std::cerr << "TravActor::leave " << this << " Node " << n 
             << " Res " << res << std::endl;
        return Action::Continue;
    }
};

int main( int argc, char *argv[] )
{
    osgInit(argc, argv);

    // build simple tree: g1|((g2|t2),t1)

    NodePtr g1 = Node::create();
    Core1Ptr g1c = Core1::create();
    g1->setCore( g1c );
    NodePtr g2 = Node::create();
    Core1Ptr g2c = Core1::create();
    g2->setCore( g2c );
    NodePtr t1 = Node::create();
    Core2Ptr t1c = Core2::create();
    t1->setCore( t1c );
    NodePtr t2 = Node::create();
    Core2Ptr t2c = Core2::create();
    t2->setCore( t2c );

    g2->addChild( t2 );

    g1->addChild( g2 );
    g1->addChild( t1 );

    // special functor cast test!
#if 0
    NodePtr lightnode = Node::create();
    DirectionalLightPtr light = DirectionalLight::create();
    lightnode->setCore(light);
    Action::Functor functor = std::bind(reinterpret_cast<Action::Callback>(&DirectionalLight::renderEnter), std::placeholders::_1, std::placeholders::_2);
    printf("light functor1\n");
    functor(light, nullptr);
    printf("light functor2\n");
    light->renderEnter(nullptr);
#endif

    // two instances for check instance calling
    Core1Action gf1,gf2;

    // register a default function. Should be copied at 
    // instantiation time

    Action::registerEnterDefault(Core1::getClassType(), 
                                 &defenter1);

    Action::registerLeaveDefault(Core1::getClassType(), 
                                 &defleave1);

    Action *act1;
    act1 = Action::create();

    std::cerr << "Apply (c2 unset):" << std::endl;
    act1->apply(g2);    

    Action::registerEnterDefault(Core2::getClassType(), 
                                 &defenter2);
    Action::registerLeaveDefault(Core2::getClassType(), 
                                 &defleave2);

    std::cerr << "Apply (c2 unset, pickup from default):" << std::endl;
    act1->apply(g2);    


    Action *act2;
    act2 = Action::create();

    std::cerr << "Apply (c2 set):" << std::endl;
    act2->apply(g2);    

    // assign functors
    act1->registerEnterFunction(Core1::getClassType(),
                                std::bind(&Core1Action::enter, &gf1, std::placeholders::_1, std::placeholders::_2));

    act1->registerLeaveFunction(Core1::getClassType(),
                                std::bind(&Core1Action::leave, &gf2, std::placeholders::_1, std::placeholders::_2));
    
    act1->registerEnterFunction(Core2::getClassType(), 
                                &Core2Enter);
    act1->registerLeaveFunction(Core2::getClassType(), 
                                &Core2Leave);

    // set act1 as the prototype
    
    Action::setPrototype( act1 );
    
    // call on single node

    std::cerr << "Apply(leaf):" << std::endl;
    act1->apply( t1 );

    // call on single node

    std::cerr << "Apply(leaf), second action :" << std::endl;
    act2->apply( t1 );
    
    // call on single node
    // act2 should now have the same functions
    
    delete act2;
    act2 = Action::create();

    std::cerr << "Apply(leaf), second action from prototype:" << std::endl;
    act2->apply( t1 );

    // call on node with single child:

    std::cerr << "Apply(node):" << std::endl;
    act1->apply( g2 );

    // call on tree:

    std::cerr << "Apply(tree):" << std::endl;
    act1->apply( g1 );

    //use a function that only traverses the first node (if any)
    act1->registerEnterFunction(Core1::getClassType(), 
                                &firstOnly);

    std::cerr << "Apply(single child traversal):" << std::endl;

    act1->apply( g1 );

    // call on node with single child:

    std::cerr << "Apply(list):" << std::endl;
    act1->apply( g1->editMFChildren()->begin(), g1->editMFChildren()->end() );

    {
    /*
    TravActor tact1,tact2;
    // traversal function test
    std::cerr << "Traverse(node,enter):" << std::endl;
    traverse(g1, std::bind(&TravActor::enter, &tact1, std::placeholders::_1));

    std::cerr << "Traverse(node,enter):" << std::endl;
    traverse(g1, std::bind(&TravActor::enter, &tact1, std::placeholders::_1));
    
    std::cerr << "Traverse(list,enter):" << std::endl;
    traverse(g1->getMFChildren()->getValues(),
             std::bind(&TravActor::enter, &tact2, std::placeholders::_1));
    
    std::cerr << "Traverse(node,enter&leave):" << std::endl;
    traverse(g1, std::bind(&TravActor::enter, &tact1, std::placeholders::_1),
             std::bind(&TravActor::leave, &tact1, std::placeholders::_1, std::placeholders::_2));

    
    std::cerr << "Traverse(list,enter&leave):" << std::endl;
    traverse(g1->getMFChildren()->getValues(), std::bind(&TravActor::enter, &tact2, std::placeholders::_1),
             std::bind(&TravActor::leave, &tact2, std::placeholders::_1, std::placeholders::_2));

    std::cerr << "Traverse(node,enter):" << std::endl;
    traverse(g1, &traventer);
    
    std::cerr << "Traverse(list,enter):" << std::endl;
    traverse(g1->getMFChildren()->getValues(), &traventer);
    
    std::cerr << "Traverse(node,enter&leave):" << std::endl;
    traverse(g1, &traventer, &travleave);
    
    std::cerr << "Traverse(list,enter&leave):" << std::endl;
    traverse(g1->getMFChildren()->getValues(), 
             &traventer, &travleave);
                 */
    }

    // try the error checks

    // NULL nodes

    /*
    std::vector<NodePtr> nullvec;
    nullvec.push_back( NullFC );

    std::cerr << "Apply(list) Null:" << std::endl;
    act1->apply( nullvec.begin(), nullvec.end() );

    std::cerr << "Apply(node) Null:" << std::endl;
    act1->apply( NullFC );

    NodePtr g3 = Node::create();
    std::cerr << "Apply(node) without core:" << std::endl;
    act1->apply( g3 );
*/
}
