/*---------------------------------------------------------------------------*\
 *                                OpenSG                                     *
 *                                                                           *
 *                                                                           *
 *                     Copyright 2000-2002 by OpenSG Forum                   *
 *                                                                           *
 *   contact: dirk@opensg.org, gerrit.voss@vossg.org, jbehr@zgdv.de          *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                License                                    *
 *                                                                           *
 * This library is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU Library General Public License as published    *
 * by the Free Software Foundation, version 2.                               *
 *                                                                           *
 * This library is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 * Library General Public License for more details.                          *
 *                                                                           *
 * You should have received a copy of the GNU Library General Public         *
 * License along with this library; if not, write to the Free Software       *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 *
 *                                                                           *
\*---------------------------------------------------------------------------*/
/*---------------------------------------------------------------------------*\
 *                                Changes                                    *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
 *                                                                           *
\*---------------------------------------------------------------------------*/

#ifdef OSG_DOC_FILES_IN_MODULE
/*! \file OSGNode.cpp
    \ingroup GrpSystemFieldContainer
 */
#endif

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

#include <functional>
#include <algorithm>

#include "OSGConfig.h"
#include "OSGFieldContainerPtr.h"
#include "OSGFieldContainerType.h"
#include "OSGNode.h"
#include "OSGNodeCore.h"
#include "OSGBinaryDataHandler.h"
#include "OSGSFFieldContainerPtr.h"
#include "OSGMFFieldContainerPtr.h"
#include "OSGName.h"

OSG_USING_NAMESPACE

const BitVector Node::VolumeFieldMask      =
        (TypeTraits<BitVector>::One << Node::VolumeFieldId     );
const BitVector Node::TravMaskFieldMask      =
        (TypeTraits<BitVector>::One << Node::TravMaskFieldId   );
const BitVector Node::ParentFieldMask      =
        (TypeTraits<BitVector>::One << Node::ParentFieldId     );
const BitVector Node::ChildrenFieldMask    =
        (TypeTraits<BitVector>::One << Node::ChildrenFieldId   );
const BitVector Node::CoreFieldMask        =
        (TypeTraits<BitVector>::One << Node::CoreFieldId       );
const BitVector Node::FlagsFieldMask        =
        (TypeTraits<BitVector>::One << Node::FlagsFieldId      );
const BitVector Node::AssociatedFieldMask  =
        (TypeTraits<BitVector>::One << Node::AssociatedFieldId );
const BitVector Node::ExternalFlagsFieldMask =
    (TypeTraits<BitVector>::One << Node::ExternalFlagsFieldId      );

FieldDescription *Node::_desc[] =
{
#ifdef OSG_DYNAMIC_VOLUME
    new FieldDescription(SFDynamicVolume::getClassType(),
                         "volume",
                         OSG_FC_FIELD_IDM_DESC(VolumeField),
                         FieldDescription::FDExternal,
                         reinterpret_cast<FieldAccessMethod>(
                             &Node::editSFVolume)),
#else
    new FieldDescription(SFBoxVolume::getClassType(),
                         "volume",
                         OSG_FC_FIELD_IDM_DESC(VolumeField),
                         FieldDescription::FDExternal,
                         reinterpret_cast<FieldAccessMethod>(
                             &Node::editSFVolume)),
#endif

    // Yes, this is wrong, it should be an UInt32, but changing
    // it now will break all old .osb files, and this info is
    // nearly never necessary to be loaded from files.
    new FieldDescription(SFBool::getClassType(),
                         "travMask",
                         OSG_FC_FIELD_IDM_DESC(TravMaskField),
                         FieldDescription::FDExternal,
                         reinterpret_cast<FieldAccessMethod>(
                             &Node::editSFTravMask)),

    new FieldDescription(SFNodePtr::getClassType(),
                         "parent",
                         OSG_FC_FIELD_IDM_DESC(ParentField),
                         FieldDescription::FDInternal,
                         reinterpret_cast<FieldAccessMethod>(
                             &Node::editSFParent)),

    new FieldDescription(MFNodePtr::getClassType(),
                         "children",
                         OSG_FC_FIELD_IDM_DESC(ChildrenField),
                         FieldDescription::FDExternal,
                         reinterpret_cast<FieldAccessMethod>(
                             &Node::editMFChildren)),

    new FieldDescription(SFNodeCorePtr::getClassType(),
                         "core",
                         OSG_FC_FIELD_IDM_DESC(CoreField),
                         FieldDescription::FDExternal,
                         reinterpret_cast<FieldAccessMethod>(&Node::editSFCore)),

    new FieldDescription(SFUInt32::getClassType(),
                         "flags",
                         OSG_FC_FIELD_IDM_DESC(FlagsField),
                         FieldDescription::FDInternal,
                         reinterpret_cast<FieldAccessMethod>(
                             &Node::editSFFlags)),

    new FieldDescription(SFNodePtr::getClassType(),
                         "associated",
                         OSG_FC_FIELD_IDM_DESC(AssociatedField),
                         FieldDescription::FDExternal,
                         reinterpret_cast<FieldAccessMethod>(
                             &Node::editSFAssociated)),

    new FieldDescription(SFUInt32::getClassType(),
                         "externalFlags",
                         OSG_FC_FIELD_IDM_DESC(ExternalFlagsField),
                         FieldDescription::FDExternal,
                         reinterpret_cast<FieldAccessMethod>(
                             &Node::editSFExternalFlags)),
};

FieldContainerType Node::_type(
    "Node",
    "AttachmentContainer",
    0,
    reinterpret_cast<PrototypeCreateF>(&Node::createEmpty),
    0,
    _desc,
    sizeof(_desc));


const NodePtr Node::NullNode(NullFC);

OSG_FIELD_CONTAINER_DEF(Node, NodePtr)

/*-------------------------------------------------------------------------*/
/*                                Set                                      */

void Node::setCore(const NodeCorePtr &core)
{
    NodePtr thisP = getPtr();

    thisP.setParentFieldPos(CoreFieldId);

    addRefCP(core);

    if(_sfCore.getValue() != NullFC)
    {
        beginEditCP(_sfCore.getValue(), NodeCore::ParentsFieldMask);
        {
            _sfCore.getValue()->subParent(thisP);
        }
        endEditCP  (_sfCore.getValue(), NodeCore::ParentsFieldMask);

        subRefCP(_sfCore.getValue());
    }

    _sfCore.setValue(core);

    if(_sfCore.getValue() != NullFC)
    {
        beginEditCP(_sfCore.getValue(), NodeCore::ParentsFieldMask);
        {
            _sfCore.getValue()->addParent(thisP);
        }
        endEditCP  (_sfCore.getValue(), NodeCore::ParentsFieldMask);
    }

    // TODO Check if required (GV)
    invalidateVolume();
}

/*-------------------------------------------------------------------------*/
/*                             Children                                    */

void Node::addChild(const NodePtr &childP)
{
    if(childP != NullFC)
    {
        // do the ref early, to prevent destroys on getParent(a)->addChild(a)
        addRefCP(childP);

        // already somebody else's child?
        NodePtr parent = childP->getParent();
        if(parent != NullFC)
        {
            beginEditCP(parent, Node::ChildrenFieldMask);
            parent->subChild(childP);
            endEditCP  (parent, Node::ChildrenFieldMask);
        }

        _mfChildren.push_back(childP);

        beginEditCP(childP, Node::ParentFieldMask);
        {
            childP->setParent(getPtr());
        }
        endEditCP  (childP, Node::ParentFieldMask);

        // TODO Check if required (GV)
#ifndef OSG_GV_BETA
        invalidateVolume();
#endif
    }
}

void Node::insertChild(UInt32 childIndex, const NodePtr &childP)
{
    MFNodePtr::iterator childIt = _mfChildren.begin();

    if(childP != NullFC)
    {
        // do the ref early, to prevent destroys on getParent(a)->addChild(a)
        addRefCP(childP);

        // already somebody else's child?
        NodePtr parent = childP->getParent();
        if(parent != NullFC)
        {
            beginEditCP(parent, Node::ChildrenFieldMask);
            parent->subChild(childP);
            endEditCP  (parent, Node::ChildrenFieldMask);
        }

        childIt += childIndex;

        _mfChildren.insert(childIt, childP);

        beginEditCP(childP, Node::ParentFieldMask);
        {
            childP->setParent(getPtr());
        }
        endEditCP  (childP, Node::ParentFieldMask);
    }

    // TODO check if required (GV)
#ifndef OSG_GV_BETA
    invalidateVolume();
#endif
}

void Node::insertChild(const NodePtr &posChildP, const NodePtr &childP, bool after)
{
    if(posChildP == childP)
        return;

    if(childP == NullFC)
        return;

    MFNodePtr::iterator childIt = _mfChildren.find(posChildP);

    if(childIt == _mfChildren.end())
        return;

    // do the ref early, to prevent destroys on getParent(a)->addChild(a)
    addRefCP(childP);

    // already somebody else's child?
    NodePtr parent = childP->getParent();
    if(parent != NullFC)
    {
        beginEditCP(parent, Node::ChildrenFieldMask);
        parent->subChild(childP);
        endEditCP  (parent, Node::ChildrenFieldMask);
    }

    // ok the subChild invalidates the childIt iterator!
    childIt = _mfChildren.find(posChildP);
    if(after)
        ++childIt;
    _mfChildren.insert(childIt, childP);
    beginEditCP(childP, Node::ParentFieldMask);
    {
        childP->setParent(getPtr());
    }
    endEditCP  (childP, Node::ParentFieldMask);
}

void Node::replaceChild(UInt32 childIndex, const NodePtr &childP)
{
    if(childP != NullFC && childIndex < _mfChildren.size())
    {
        // do the ref early, to prevent destroys on getParent(a)->addChild(a)
        addRefCP(childP);
        // already somebody else's child?
        // moved it up could be a child of childIndex.
        NodePtr parent = childP->getParent();
        if(parent != NullFC)
        {
            beginEditCP(parent, Node::ChildrenFieldMask);
            parent->subChild(childP);
            endEditCP  (parent, Node::ChildrenFieldMask);
        }

        // remove the current child
        beginEditCP(_mfChildren[childIndex], Node::ParentFieldMask);
        {
            _mfChildren[childIndex]->setParent(NullNode);
        }
        endEditCP  (_mfChildren[childIndex], Node::ParentFieldMask);

        subRefCP(_mfChildren[childIndex]);

        // set the new child
        _mfChildren[childIndex] = childP;

        beginEditCP(childP, Node::ParentFieldMask);
        {
            childP->setParent(getPtr());
        }
        endEditCP  (childP, Node::ParentFieldMask);
    }

    // TODO check if required (GV)
#ifndef OSG_GV_BETA
    invalidateVolume();
#endif
}

//! return true on success, false on child not found

bool Node::replaceChildBy(const NodePtr &childP,
                          const NodePtr &newChildP)
{
    MFNodePtr::iterator childIt = _mfChildren.find(childP);

    if(newChildP != NullFC)
    {
        if(childIt != _mfChildren.end())
        {
            // do the ref early, to prevent destroys on
            // getParent(a)->addChild(a)

            addRefCP(newChildP);
            // already somebody else's child?
            // moved it up could be a child of childP.
            NodePtr parent = newChildP->getParent();
            if(parent != NullFC)
            {
                beginEditCP(parent, Node::ChildrenFieldMask);
                parent->subChild(newChildP);
                endEditCP  (parent, Node::ChildrenFieldMask);
            }

            beginEditCP(childP, Node::ParentFieldMask);
            {
                childP->setParent(NullNode);
            }
            endEditCP  (childP, Node::ParentFieldMask);

            subRefCP(childP);

            (*childIt) = newChildP;

            beginEditCP(newChildP, Node::ParentFieldMask);
            {
                newChildP->setParent(getPtr());
            }
            endEditCP  (newChildP, Node::ParentFieldMask);

            // TODO check if required (GV)
#ifndef OSG_GV_BETA
            invalidateVolume();
#endif

            return true;
        }
    }

    return false;
}

Int32 Node::findChild(const NodePtr &childP) const
{
    UInt32 index;

    for(index = 0; index < _mfChildren.size(); index++)
    {
        if( _mfChildren[index] == childP)
            break;
    }

    if(index < _mfChildren.size())
        return index;
    else
        return -1;
}

void Node::subChild(const NodePtr &childP)
{
    MFNodePtr::iterator childIt = _mfChildren.find(childP);

    if(childIt != _mfChildren.end())
    {
        beginEditCP(childP, Node::ParentFieldMask);
        {
            childP->setParent(NullNode);
        }
        endEditCP  (childP, Node::ParentFieldMask);

        subRefCP(childP);

        _mfChildren.erase(childIt);
    }
    else
    {
        SWARNING << "Node(" << this << ")::subChild: " << childP
                 << " is not one of my children!" << std::endl;
    }

    // TODO check if required (GV)
#ifndef OSG_GV_BETA
    invalidateVolume();
#endif
}

void Node::subChild(UInt32 childIndex)
{
    MFNodePtr::iterator childIt = _mfChildren.begin();

    childIt += childIndex;

    if(childIt != _mfChildren.end())
    {
        beginEditCP(*childIt, Node::ParentFieldMask);
        {
            (*childIt)->setParent(NullNode);
        }
        endEditCP  (*childIt, Node::ParentFieldMask);

        subRefCP(*childIt);

        _mfChildren.erase(childIt);
    }

    // TODO check if required (GV)
#ifndef OSG_GV_BETA
    invalidateVolume();
#endif
}


/*-------------------------------------------------------------------------*/
/*                           Get Transformation                            */

Matrix Node::getToWorld(void)
{
    Matrix tmp;

    getToWorld(tmp);

    return tmp;
}

void Node::getToWorld(Matrix &result)
{
    NodePtr parent = getParent();
    if(parent != NullFC)
    {
        parent->getToWorld(result);
    }
    else
    {
        result.setIdentity();
    }
    NodeCorePtr core = getCore();
    if(core != NullFC)
        core->accumulateMatrix(result);
}

/*-------------------------------------------------------------------------*/
/*                           Volume                                        */

#ifdef OSG_DYNAMIC_VOLUME
void Node::getWorldVolume(DynamicVolume &result)
#else
void Node::getWorldVolume(BoxVolume &result)
#endif
{
    Matrix m;
    NodePtr parent = getParent();
    if(parent != NullFC)
    {
        parent->getToWorld(m);
    }
    else
    {
        m.setIdentity();
    }

    updateVolume();

    result = getVolume();
    result.transform(m);
/*
Pnt3f low,high;
result.getBounds(low,high);
fprintf(stderr,"%p: node 0x%p gwv (%f %f %f  %f %f %f)\n",
            Thread::getCurrent(), this,
            low[0], low[1], low[2],
            high[0], high[1], high[2] );
*/
}

void Node::updateVolume(void)
{
#ifdef OSG_DYNAMIC_VOLUME
    if(_sfVolume.getValue().getInstance().isValid() == true ||
       getTravMask()                                == 0x0000)
    {
        return;             // still valid, nothing to do
    }

    // be careful to not change the real volume. If two threads
    // are updating the same aspect this will lead to chaos
    DynamicVolume vol = _sfVolume.getValue();

//fprintf(stderr,"%p: node 0x%p update needed\n", Thread::getCurrent(), this);

    MFNodePtr::iterator it;

    vol.getInstance().setEmpty();

    for(it = _mfChildren.begin(); it != _mfChildren.end(); ++it)
    {
        if(*it != NullFC && (*it)->getTravMask())
        {
            (*it)->updateVolume();
            vol.getInstance().extendBy((*it)->getVolume());
        }
    }

    // test for null core. Shouldn't happen, but just in case...
    if(getCore() != NullFC)
        getCore()->adjustVolume(vol.getInstance());

    NodePtr thisP = getPtr();

    beginEditCP(thisP, VolumeFieldMask);

    vol.instanceChanged();

    _sfVolume.setValue(vol);

    endEditCP(thisP, VolumeFieldMask);

#else

    if(_sfVolume.getValue().isValid() == true ||
       getTravMask()                  == 0x0000)
    {
        return;             // still valid, nothing to do
    }

    // be careful to not change the real volume. If two threads
    // are updating the same aspect this will lead to chaos
    BoxVolume vol = _sfVolume.getValue();

//fprintf(stderr,"%p: node 0x%p update needed\n", Thread::getCurrent(), this);

    MFNodePtr::iterator it;

    vol.setEmpty();

    for(it = _mfChildren.begin(); it != _mfChildren.end(); ++it)
    {
        if(*it != NullFC && (*it)->getTravMask())
        {
            (*it)->updateVolume();
            vol.extendBy((*it)->getVolume());
        }
    }

    // test for null core. Shouldn't happen, but just in case...
    if(getCore() != NullFC)
        getCore()->adjustVolume(vol);

    NodePtr thisP = getPtr();

    beginEditCP(thisP, VolumeFieldMask);

    _sfVolume.setValue(vol);

    endEditCP(thisP, VolumeFieldMask);
#endif // OSG_2_PREP
}

void Node::invalidateVolume(void)
{
#ifdef OSG_DYNAMIC_VOLUME
    Volume &vol = _sfVolume.getValue().getInstance();
#else
    Volume &vol = _sfVolume.getValue();
#endif

    if(vol.isValid() == true && vol.isStatic() == false)
    {
        NodePtr thisP = getPtr();

        beginEditCP(thisP, VolumeFieldMask);

        vol.setValid(false);
#ifdef OSG_DYNAMIC_VOLUME
        _sfVolume.getValue().instanceChanged();
#endif

        endEditCP(thisP, VolumeFieldMask);

        if(getParent() != NullFC)
        {
            getParent()->invalidateVolume();
        }
    }
}

/*-------------------------------------------------------------------------*/
/*                                Dump                                     */

void Node::dump(      UInt32    uiIndent,
                const BitVector bvFlags) const
{
    UInt32 i;

    NodePtr thisP = getPtr();

    indentLog(uiIndent, PLOG);

    PLOG << "Node"
         << "("
         << getContainerId(thisP)
         << ") : "
         << _mfChildren.size()
         << " children | "
         << _attachmentMap.getValue().size()
         << " attachments | "
         << "Parent : " << std::hex;

    if(_sfParent.getValue() != NullFC)
        PLOG << "0x" << &(*(_sfParent.getValue())) << " | ";
    else
        PLOG << "NULL | ";

    PLOG << "0x" << this << std::dec << std::endl;

    indentLog(uiIndent, PLOG);

    PLOG << "[" << std::endl;

    if(_sfCore.getValue() != NullFC)
    {
        _sfCore.getValue()->dump(uiIndent + 4, bvFlags);
    }
    else
    {
        indentLog(uiIndent + 4, PLOG);
        PLOG << "Core : " << "NULL" << std::endl;
    }

    Inherited::dump(uiIndent, bvFlags);

    indentLog(uiIndent, PLOG);
    PLOG << "]" << std::endl;

    indentLog(uiIndent, PLOG);

    PLOG << "{" << std::endl;

    for(i = 0; i < _mfChildren.size(); i++)
    {
        _mfChildren[i]->dump(uiIndent + 4, bvFlags);
        PLOG << std::endl;
    }


    indentLog(uiIndent, PLOG);

    PLOG << "}" << std::endl;

/*
    for(i = 0; i < indent; i++)
    {
        cerr << " ";
    }
    cerr << "  Parent : ";


    for(i = 0; i < indent; i++)
    {
        cerr << " ";
    }

    for(i = 0; i < indent; i++)
        fprintf(stderr, " ");

    fprintf(stderr, "NAttachments : \n");

    map<UInt32, AttachmentPtr>::const_iterator fcI;

    fcI = _attachmentMap.getValue().begin();

    while(fcI != _attachmentMap.getValue().end())
    {
        (*fcI).second->dump(indent + 2);
        ++fcI;
    }

    for(i = 0; i < indent; i++)
    {
        cerr << " ";
    }

    cerr << "{" << std::endl;


    for(i = 0; i < indent; i++)
    {
        cerr << " ";
    }

    cerr << "}" << std::endl;
*/
}

/*-------------------------------------------------------------------------*/
/*                            Constructors                                 */

Node::Node(void) :
     Inherited    (),
    _sfVolume     (),
    _sfTravMask   (TypeTraits<UInt32>::getMax()),
    _sfParent     (),
    _mfChildren   (),
    _sfCore       (),
    _sfFlags      (0),
    _occlusionMask(0),
    _sfAssociated (),
    _sfExternalFlags(0)
{
}

Node::Node(const Node &source) :
     Inherited       (source),
    _sfVolume        (source._sfVolume),
    _sfTravMask      (source._sfTravMask),
    _sfParent        (),
    _mfChildren      (),
    _sfCore          (),
    _sfFlags         (source._sfFlags),
    _occlusionMask   (source._occlusionMask),
    _sfAssociated    (),
    _sfExternalFlags(source._sfExternalFlags)
{
}

/*-------------------------------------------------------------------------*/
/*                             Destructor                                  */

Node::~Node(void)
{
    if(_sfCore.getValue() != NullFC)
    {
        NodePtr thisP = getPtr();

        beginEditCP(_sfCore.getValue(), NodeCore::ParentsFieldMask);
        {
            _sfCore.getValue()->subParent(thisP);
        }
        endEditCP  (_sfCore.getValue(), NodeCore::ParentsFieldMask);

        subRefCP(_sfCore.getValue());
    }

    MFNodePtr::iterator       vChildIt    = _mfChildren.begin();
    MFNodePtr::const_iterator endChildren = _mfChildren.end  ();

    while(vChildIt != endChildren)
    {
        beginEditCP(*vChildIt, Node::ParentFieldMask);
        {
            (*vChildIt)->setParent(NullNode);
        }
        endEditCP  (*vChildIt, Node::ParentFieldMask);

        subRefCP(*vChildIt);

        ++vChildIt;
    }
}

#if defined(OSG_FIXED_MFIELDSYNC)
void Node::onDestroyAspect(UInt32 uiId, UInt32 uiAspect)
{
    _mfChildren.terminateShare(uiAspect, this->getContainerSize());
}
#endif

void osg::addAttachments(const AttachmentContainerPtr &src,
    AttachmentContainerPtr dst)
{
    if(src == NullFC || dst == NullFC)
        return;

    // add all attachments.
    SFAttachmentMap *amap = static_cast<SFAttachmentMap *>(src->getSFAttachments());
    AttachmentMap::const_iterator   mapIt = amap->getValue().begin();
    AttachmentMap::const_iterator   mapEnd = amap->getValue().end();

    beginEditCP(dst, AttachmentContainer::AttachmentsFieldMask);
    for(; mapIt != mapEnd; ++mapIt)
    {
        if(mapIt->second != NullFC)
            dst->addAttachment(AttachmentPtr::dcast(mapIt->second));
    }
    endEditCP  (dst, AttachmentContainer::AttachmentsFieldMask);
}

NodePtr OSG::cloneTree(const NodePtr &pRootNode, bool cloneAttachments)
{
    NodePtr returnValue = NullFC;

    if(pRootNode != NullFC)
    {
        NodePtr pChildClone = NullFC;

        returnValue = Node::create();
        if(cloneAttachments)
        {
            OSG::addAttachments(pRootNode, returnValue);
        }
        else
        {
            // copy name
            OSG::copyName(pRootNode, returnValue);
        }

        beginEditCP(returnValue);
        {
            returnValue->setTravMask(pRootNode->getTravMask());
            returnValue->setCore    (pRootNode->getCore());
            returnValue->setExternalFlags(pRootNode->getExternalFlags());

            for(UInt32 i = 0; i < pRootNode->getNChildren(); i++)
            {
                pChildClone = cloneTree(pRootNode->getChild(i), cloneAttachments);
                returnValue->addChild(pChildClone);
            }
        }
        endEditCP  (returnValue);
    }

    return returnValue;
}

// we need to do this for the light sources beacon field!
static void fixParents(const NodePtr &node)
{
    if(node == NullFC)
        return;

    NodeCorePtr core = NodeCorePtr::dcast(node->getCore());
    if(core != NullFC)
    {
        beginEditCP(node);
            node->setCore(core);
        endEditCP(node);
    }

    beginEditCP(node);
        std::vector<NodePtr> childs;
        for(UInt32 i=0;i<node->getNChildren();++i)
        {
            NodePtr child = node->getChild(i);
            addRefCP(child);
            childs.push_back(child);
            node->subChild(child);
        }
        for(UInt32 i=0;i<childs.size();++i)
            node->addChild(childs[i]);
    endEditCP(node);

    for(UInt32 i=0;i<node->getNChildren();++i)
        fixParents(node->getChild(i));
}

// deep clone of a fieldcontainer.
// we should move this into OSGFieldContainer.cpp?
static FieldContainerPtr deepCloneR(std::map<FieldContainerPtr, FieldContainerPtr> &cloned,
                                    const FieldContainerPtr &src, const std::vector<std::string> &share)
{
    if(src == NullFC)
        return NullFC;

    const FieldContainerType &type = src->getType();

    FieldContainerType *geoType = FieldContainerFactory::the()->findType("Geometry");
    const bool isGeometry = geoType != NULL && type.isDerivedFrom(*geoType);

    const int numGeoFieldTypes = 8;
    FieldContainerType *geoFieldTypes[numGeoFieldTypes] = {0, 0, 0, 0, 0, 0, 0, 0};
    if(isGeometry)
    {
        geoFieldTypes[0] = FieldContainerFactory::the()->findType("Transform3D");
        geoFieldTypes[1] = FieldContainerFactory::the()->findType("GeoPTypes");
        geoFieldTypes[2] = FieldContainerFactory::the()->findType("GeoPLengths");
        geoFieldTypes[3] = FieldContainerFactory::the()->findType("GeoPositions");
        geoFieldTypes[4] = FieldContainerFactory::the()->findType("GeoNormals");
        geoFieldTypes[5] = FieldContainerFactory::the()->findType("GeoColors");
        geoFieldTypes[6] = FieldContainerFactory::the()->findType("GeoTexCoords");
        geoFieldTypes[7] = FieldContainerFactory::the()->findType("GeoIndices");
    }

    //FDEBUG(("deepClone: fieldcontainertype = %s\n", type.getCName()));

    FieldContainerPtr dst = FieldContainerFactory::the()->createFieldContainer(type.getName().str());

    cloned.insert(std::make_pair(src, dst));

    //UInt32 fcount = type.getNumFieldDescs();
    // ignore dynamic fields.
    UInt32 fcount = osgMin(type.getNumFieldDescs(), dst->getType().getNumFieldDescs());


    for(UInt32 i=1;i <= fcount;++i)
    {
        const FieldDescription* fdesc = type.getFieldDescription(i);

        if(fdesc->isInternal())
            continue;

        BitVector mask = fdesc->getFieldMask();

        Field *src_field = src->getField(i);
        Field *dst_field = dst->getField(i);

        const FieldType &ftype = src_field->getType();

        std::string fieldType = ftype.getName().str();

        // ignore the parents field, would lead to a endless recursion
        // with the new image parents support!
        if(strcmp(fdesc->getCName(), "parents") == 0)
            continue;

        // attachements
        if(strcmp(fdesc->getCName(), "attachments") == 0)
        {
            SFAttachmentMap *amap = static_cast<SFAttachmentMap *>(src_field);

            AttachmentMap::const_iterator   mapIt = amap->getValue().begin();
            AttachmentMap::const_iterator   mapEnd = amap->getValue().end();

            beginEditCP(dst, mask);
            for(; mapIt != mapEnd; ++mapIt)
            {
                FieldContainerPtr fc = mapIt->second;

                bool shareit = false;
                for(UInt32 k=0;k<share.size();++k)
                {
                    FieldContainerType *fct = FieldContainerFactory::the()
                                            ->findType(share[k].c_str());
                    if(fc != NullFC && fct != NULL &&
                       fc->getType().isDerivedFrom(*fct))
                    {
                        shareit = true;
                        break;
                    }
                }

                if(!shareit)
                {
                    std::map<FieldContainerPtr, FieldContainerPtr>::iterator cit = cloned.find(fc);
                    if(cit != cloned.end())
                    {
                        fc = (*cit).second;
                    }
                    else
                    {
                        fc = deepCloneR(cloned, fc, share);
                        if(fc != NullFC && fc->getType().isDerivedFrom(Node::getClassType()))
                            fixParents(NodePtr::dcast(fc));
                    }
                }

                if(fc != NullFC)
                    AttachmentContainerPtr::dcast(dst)->addAttachment(
                        AttachmentPtr::dcast(fc));
            }
            endEditCP(dst, mask);
            continue;
        }

        // field
        if(strstr(ftype.getCName(), "Ptr") == NULL)
        {
            //FDEBUG(("deepClone: fieldname = %s fieldType = %s\n", fdesc->getCName(), fieldType.c_str()));
            beginEditCP(dst, mask);
                dst_field->setAbstrValue(*src_field);
            endEditCP(dst, mask);
        }
        else // field with pointer
        {
            if(src_field->getCardinality() == FieldType::SINGLE_FIELD)
            {
                FieldContainerPtr fc =
                    (static_cast<SFFieldContainerPtr *>(src_field)->getValue());

                bool shareit = false;
                for(UInt32 k=0;k<share.size();++k)
                {
                    FieldContainerType *fct = FieldContainerFactory::the()
                                            ->findType(share[k].c_str());
                    if(fc != NullFC && fct != NULL &&
                       fc->getType().isDerivedFrom(*fct))
                    {
                        shareit = true;
                        break;
                    }
                }

                if(!shareit)
                {
                    std::map<FieldContainerPtr, FieldContainerPtr>::iterator cit = cloned.find(fc);
                    bool reuseCloned = cit != cloned.end();
                    if(reuseCloned && isGeometry && fc != NullFC)
                    {
                        for(int i = 0; i < numGeoFieldTypes; i++)
                        {
                            if( geoFieldTypes[i] != NULL && fc->getType().isDerivedFrom(*geoFieldTypes[i]))
                            {
                                reuseCloned = false;
                                break;
                            }
                        }
                    }

                    if(reuseCloned)
                    {
                        fc = (*cit).second;
                    }
                    else
                    {
                        fc = deepCloneR(cloned, fc, share);
                        // we need to do this for the light sources beacon field!
                        if(fc != NullFC && fc->getType().isDerivedFrom(Node::getClassType()))
                            fixParents(NodePtr::dcast(fc));
                    }
                }

                if(fc != NullFC)
                {
                    SFFieldContainerPtr *sfcp_field = static_cast<SFFieldContainerPtr *>(dst_field);
                    if(sfcp_field->getValue() != NullFC) // could be created in onCreate()
                        subRefCP(sfcp_field->getValue());
                    // increment reference counter!
                    addRefCP(fc);
                    beginEditCP(dst);
                        sfcp_field->setValue(fc);
                    endEditCP(dst);
                }
            }
            else if(src_field->getCardinality() == FieldType::MULTI_FIELD)
            {
                beginEditCP(dst, mask);
                for(UInt32 j=0;j < (static_cast<MFFieldContainerPtr *>(src_field))->size();++j)
                    {
                        FieldContainerPtr fc = (*((static_cast<MFFieldContainerPtr *>(src_field))))[j];

                        bool shareit = false;
                        for(UInt32 k=0;k<share.size();++k)
                        {
                            FieldContainerType *fct = FieldContainerFactory::the()
                                            ->findType(share[k].c_str());
                            if(fc != NullFC && fct != NULL &&
                               fc->getType().isDerivedFrom(*fct))
                            {
                                shareit = true;
                                break;
                            }
                        }

                        if(!shareit)
                        {
                            std::map<FieldContainerPtr, FieldContainerPtr>::iterator cit = cloned.find(fc);
                            if(cit != cloned.end())
                            {
                                fc = (*cit).second;
                            }
                            else
                            {
                                fc = deepCloneR(cloned, fc, share);
                                if(fc != NullFC && fc->getType().isDerivedFrom(Node::getClassType()))
                                    fixParents(NodePtr::dcast(fc));
                            }
                        }

                        if(fc != NullFC)
                        {
                            // increment reference counter!
                            addRefCP(fc);
                            (static_cast<MFFieldContainerPtr *>(dst_field))->push_back(fc);
                        }
                    }
                endEditCP(dst, mask);
            }
        }
    }
    return dst;
}

std::vector<std::string> getCommaSeparatedNames(const std::string& str)
{
    std::vector<std::string> result;

    // parse comma separated names.
    std::string::const_iterator nextComma;
    std::string::const_iterator curPos = str.begin();
    while(curPos < str.end())
    {
        nextComma = std::find(curPos, str.end(), ',');
        // strip leading spaces
        curPos = std::find_if(curPos, nextComma, std::not1(std::ptr_fun(isspace)));
        result.push_back(std::string(curPos, nextComma));
        if(nextComma != str.end())
            curPos = ++nextComma;
        else
            break;
    }

    return result;
}

// deep clone of a fieldcontainer.
// we should move this into OSGFieldContainer.cpp?
FieldContainerPtr OSG::deepClone(const FieldContainerPtr &src,
                                 const std::vector<std::string> &share)
{
    std::map<FieldContainerPtr, FieldContainerPtr> cloned;
    return deepCloneR(cloned, src, share);
}

FieldContainerPtr OSG::deepClone(const FieldContainerPtr &src,
                                 const std::vector<UInt16> &shareGroupIds)
{
    std::vector<std::string> share;
    share.reserve(shareGroupIds.size());
    for(UInt32 i=0;i<shareGroupIds.size();++i)
    {
        const char *name = FieldContainerFactory::the()->findGroupName(shareGroupIds[i]);
        if(name != NULL)
            share.push_back(name);
    }
    return OSG::deepClone(src, share);
}

// shareString is a comma separated FieldContainer type list
// e.g. "Material, Geometry"
FieldContainerPtr OSG::deepClone(const FieldContainerPtr &src,
                                 const std::string &shareString)
{
    std::vector<std::string> share = getCommaSeparatedNames(shareString);
    return OSG::deepClone(src, share);
}

// deep clone of attachements.
static void deepCloneAttachmentsR(std::map<FieldContainerPtr, FieldContainerPtr> &cloned,
                                  const AttachmentContainerPtr &src,
                                  AttachmentContainerPtr dst,
                                  const std::vector<std::string> &share)
{
    SFAttachmentMap *amap =
        static_cast<SFAttachmentMap *>(src->getSFAttachments());

    AttachmentMap::const_iterator   mapIt = amap->getValue().begin();
    AttachmentMap::const_iterator   mapEnd = amap->getValue().end();

    beginEditCP(dst, AttachmentContainer::AttachmentsFieldMask);
    for(; mapIt != mapEnd; ++mapIt)
    {
        FieldContainerPtr fc = mapIt->second;

        bool shareit = false;
        for(UInt32 k=0;k<share.size();++k)
        {
            FieldContainerType *fct = FieldContainerFactory::the()
                                      ->findType(share[k].c_str());
            if(fc != NullFC && fct != NULL &&
               fc->getType().isDerivedFrom(*fct))
            {
                shareit = true;
                break;
            }
        }

        if(!shareit)
        {
            std::map<FieldContainerPtr, FieldContainerPtr>::iterator cit = cloned.find(fc);
            if(cit != cloned.end())
            {
                fc = (*cit).second;
            }
            else
            {
                fc = deepCloneR(cloned, fc, share);
                if(fc != NullFC && fc->getType().isDerivedFrom(Node::getClassType()))
                    fixParents(NodePtr::dcast(fc));
            }
        }

        if(fc != NullFC)
            dst->addAttachment(AttachmentPtr::dcast(fc));
    }
    endEditCP(dst, AttachmentContainer::AttachmentsFieldMask);
}

// deep clone of attachements.
void OSG::deepCloneAttachments(const AttachmentContainerPtr &src,
                               AttachmentContainerPtr dst,
                               const std::vector<std::string> &share)
{
    std::map<FieldContainerPtr, FieldContainerPtr> cloned;
    deepCloneAttachmentsR(cloned, src, dst, share);
}

void OSG::deepCloneAttachments(const AttachmentContainerPtr &src,
                               AttachmentContainerPtr dst,
                               const std::vector<UInt16> &shareGroupIds)
{
    std::vector<std::string> share;
    share.reserve(shareGroupIds.size());
    for(UInt32 i=0;i<shareGroupIds.size();++i)
    {
        const char *name = FieldContainerFactory::the()
                           ->findGroupName(shareGroupIds[i]);
        if(name != NULL)
            share.push_back(name);
    }
    OSG::deepCloneAttachments(src, dst, share);
}

// shareString is a comma separated FieldContainer type list
// e.g. "Material, Geometry"
void OSG::deepCloneAttachments(const NodePtr &src, NodePtr &dst,
                               const std::string &shareString)
{
    std::vector<std::string> share = getCommaSeparatedNames(shareString);
    OSG::deepCloneAttachments(src, dst, share);
}

// deep clone of nodes.
static NodePtr deepCloneTreeR(std::map<FieldContainerPtr, FieldContainerPtr> &cloned,
                              const NodePtr &src,
                              const std::vector<std::string> &share)
{
    NodePtr dst = NullFC;

    if(src != NullFC)
    {
        dst = Node::create();
        cloned.insert(std::make_pair(src, dst));
        deepCloneAttachmentsR(cloned, src, dst, share);

        beginEditCP(dst);
        {
            dst->setActive(src->getActive());
            dst->setTravMask(src->getTravMask());
            dst->setExternalFlags(src->getExternalFlags());
            dst->setCore(NodeCorePtr::dcast(deepCloneR(cloned, src->getCore(), share)));

            for(UInt32 i = 0; i < src->getNChildren(); i++)
                dst->addChild(deepCloneTreeR(cloned, src->getChild(i), share));
        }
        endEditCP  (dst);
    }

    return dst;
}

// deep clone of nodes.
NodePtr OSG::deepCloneTree(const NodePtr &src,
                           const std::vector<std::string> &share)
{
    std::map<FieldContainerPtr, FieldContainerPtr> cloned;
    return deepCloneTreeR(cloned, src, share);
}

NodePtr OSG::deepCloneTree(const NodePtr &src,
                           const std::vector<UInt16> &shareGroupIds)
{
    std::vector<std::string> share;
    share.reserve(shareGroupIds.size());
    for(UInt32 i=0;i<shareGroupIds.size();++i)
    {
        const char *name = FieldContainerFactory::the()->findGroupName(shareGroupIds[i]);
        if(name != NULL)
            share.push_back(name);
    }
    return OSG::deepCloneTree(src, share);
}

// shareString is a comma separated FieldContainer type list
// e.g. "Material, Geometry"
NodePtr OSG::deepCloneTree(const NodePtr &src,
                           const std::string &shareString)
{
    std::vector<std::string> share = getCommaSeparatedNames(shareString);
    return OSG::deepCloneTree(src, share);
}


static bool hasExcludedParent(const NodePtr &src, const std::set<osg::NodeCorePtr>& excludedNodeCores)
{
    if(src == NullFC)
        return false;

    NodePtr parent = src->getParent();
    while(parent != NullFC)
    {
        NodeCorePtr parentCore = parent->getCore();
        if(excludedNodeCores.find(parentCore) != excludedNodeCores.end())
            return true; // parent with excluded core found

        parent = parent->getParent();
    }

    return false;
}

static NodePtr deepCloneTreeKeepSharedR(std::map<FieldContainerPtr, FieldContainerPtr> &cloned,
    std::map<NodeCorePtr, NodeCorePtr> &clonedCores,
    const NodePtr &src,
    const std::vector<std::string> &share,
    const std::set<osg::NodeCorePtr>& excludedNodeCores)
{
    NodePtr dst = NullFC;

    if(src != NullFC)
    {
        NodeCorePtr srcCore = src->getCore();

        dst = Node::create();
        cloned.insert(std::make_pair(src, dst));

        beginEditCP(dst);
        {
            dst->setActive(src->getActive());
            dst->setTravMask(src->getTravMask());
            dst->setExternalFlags(src->getExternalFlags());

            bool isExcluded = excludedNodeCores.find(srcCore) != excludedNodeCores.end();
            if(isExcluded || hasExcludedParent(src, excludedNodeCores))
            {
                dst->setCore(srcCore);
                OSG::addAttachments(src, dst);
            }
            else
            {
                deepCloneAttachmentsR(cloned, src, dst, share);

                std::map<NodeCorePtr, NodeCorePtr>::const_iterator cit = clonedCores.find(srcCore);
                if(cit != clonedCores.end())
                {
                    dst->setCore(cit->second);
                }
                else
                {
                    osg::NodeCorePtr clonedCore = NodeCorePtr::dcast(deepCloneR(cloned, srcCore, share));
                    dst->setCore(clonedCore);
                    clonedCores.insert(std::make_pair(srcCore, clonedCore));
                }
            }

            for(UInt32 i = 0; i < src->getNChildren(); i++)
                dst->addChild(deepCloneTreeKeepSharedR(cloned, clonedCores, src->getChild(i), share, excludedNodeCores));
        }
        endEditCP  (dst);
    }

    return dst;
}

NodePtr OSG::deepCloneTreeKeepShared(const NodePtr &src,
    const std::vector<std::string> &share,
    const std::set<osg::NodeCorePtr>& excludedNodeCores)
{
    // stores all deep cloned fieldcontainers (source container -> cloned container)
    std::map<FieldContainerPtr, FieldContainerPtr> cloned;
    // stores all deep cloned cores and source core (source core -> cloned core)
    std::map<NodeCorePtr, NodeCorePtr> clonedCores;
    return deepCloneTreeKeepSharedR(cloned, clonedCores, src, share, excludedNodeCores);
}

/*! \brief Special behaviour for duplicating a node that has shared nodes
*          in its subtree.
*
*  See VRED-3177.
*  The duplicate's children will have shared nodes but will have 
*  no connection to the original (compare to Maya "Duplicate Input Graph")
*
*  \param shareString is a comma separated FieldContainer type list
*         e.g. "Material, Geometry"
*/
NodePtr OSG::deepCloneTreeKeepShared(const NodePtr &src, 
                                     const std::string &shareString,
                                     const std::set<osg::NodeCorePtr>& excludedNodeCores)
{
    std::vector<std::string> share = getCommaSeparatedNames(shareString);
    return deepCloneTreeKeepShared(src, share, excludedNodeCores);
}

/*-------------------------------------------------------------------------*/
/*                              cvs id's                                   */

#ifdef __sgi
#pragma set woff 1174
#endif

#ifdef OSG_LINUX_ICC
#pragma warning( disable : 177 )
#endif

namespace
{
    static Char8 cvsid_cpp[] = "@(#)$Id: $";
    static Char8 cvsid_hpp[] = OSGNODE_HEADER_CVSID;
    static Char8 cvsid_inl[] = OSGNODE_INLINE_CVSID;
}
